{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Basic pipes 
-}
{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE ScopedTypeVariables #-}

import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Data.Void -- from the 'void' package

infixl 5 >->
infixr 6 >~>

newtype Pipe a b r = Pipe { unPipe :: IO (PipeStep a b r) }

data PipeStep a b r =
    Pure r
  | Request (a -> Pipe a b r)
  | Respond b (Pipe a b r) 

instance Monad (Pipe a b) where
  return x = Pipe $ return (Pure x)
  x >>= f  = Pipe $ do xstep <- unPipe x
                       case xstep of
                         Request k   -> return $ Request (k >=> f)
                         Respond b k -> return $ Respond b (k >>= f)
                         Pure r      -> unPipe (f r)

instance MonadIO (Pipe a b) where 
  liftIO io = Pipe $ Pure <$> io

respond :: b -> Pipe a b ()
respond b = Pipe . return $ Respond b (return ())

request :: Pipe r b r
request = Pipe . return $ Request return 

-- "Pull based composition"
(>->) :: forall a b c r. Pipe a b r -> Pipe b c r -> Pipe a c r
(>->) = goRight
  where
    goRight :: Pipe a b r -> Pipe b c r -> Pipe a c r
    goRight p q = Pipe $ do
      qstep <- unPipe q
      case qstep of
        Respond b k -> return $ Respond b (goRight p k)
        Pure r      -> return $ Pure r
        Request k   -> unPipe $ goLeft p k

    goLeft :: Pipe a b r -> (b -> Pipe b c r) -> Pipe a c r
    goLeft p q = Pipe $ do
      pstep <- unPipe p
      case pstep of
        Request k   -> return $ Request (\a -> goLeft (k a) q)
        Pure r      -> return $ Pure r
        Respond b k -> unPipe $ goRight k (q b)
                      
-- "Push based composition"
(>~>) :: forall a b c r. Pipe a b r -> Pipe b c r -> Pipe a c r
(>~>) = goLeft
  where
    goLeft :: Pipe a b r -> Pipe b c r -> Pipe a c r
    goLeft p q = Pipe $ do
      pstep <- unPipe p
      case pstep of
        Request k   -> return $ Request (\a -> goLeft (k a) q)
        Pure r      -> return (Pure r)
        Respond b k -> unPipe $ goRight b k q
    
    goRight :: b -> Pipe a b r -> Pipe b c r -> Pipe a c r 
    goRight b p q = Pipe $ do
      qstep <- unPipe q
      case qstep of
        Respond c k -> return $ Respond c (goRight b p k) 
        Pure r      -> return (Pure r)
        Request k   -> unPipe $ goLeft p (k b)

runPipe :: Pipe () Void r -> IO r 
runPipe p = do
  step <- unPipe p
  case step of
    Pure r      -> return r
    Request k   -> runPipe (k ()) 
    Respond b _ -> absurd b

prompter :: Pipe a Int r 
prompter = do 
  liftIO $ putStrLn "Hi!"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: Pipe Int b r
printer = forever $ do
  i <- request
  liftIO $ print i 

main1 :: IO ()
main1 = runPipe (prompter >-> printer)

mapPipe :: (a -> b) -> Pipe a b r
mapPipe f = forever $ do
  a <- request
  respond (f a)

main2 :: IO ()
main2 = runPipe (prompter >-> mapPipe (* 2) >-> printer)

takePipe :: Int -> Pipe a a () 
takePipe n = replicateM_ n $ do
  a <- request
  respond a

main3 :: IO ()
main3 = runPipe (prompter >-> takePipe 3 >-> printer)
