{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Finalizers 
-}
{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE ScopedTypeVariables #-}

import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Data.Void -- from the 'void' package

-- infixl 5 >->

newtype Pipe a b r = Pipe { unPipe :: IO (PipeStep a b r) }

type Finalizer = IO ()

data PipeStep a b r =
    Pure r
  | Request Finalizer (a -> Pipe a b r)
  | Respond Finalizer b (Pipe a b r) 

instance Monad (Pipe a b) where
  return x = Pipe $ return (Pure x)
  x >>= f  = Pipe $ do 
    xstep <- unPipe x
    case xstep of
      Request z k   -> return $ Request z (k >=> f)
      Respond z b k -> return $ Respond z b (k >>= f)
      Pure r        -> unPipe (f r)

instance MonadIO (Pipe a b) where 
  liftIO io = Pipe $ Pure <$> io

respond :: b -> Pipe a b ()
respond b = Pipe . return $ Respond (return ()) b (return ())

request :: Pipe r b r
request = Pipe . return $ Request (return ()) return 

-- "Pull based composition"
(>->) :: forall a b c r. Pipe a b r -> Pipe b c r -> Pipe a c r
(>->) = goRight (return ())
 where
  goRight :: Finalizer -> Pipe a b r -> Pipe b c r -> Pipe a c r
  goRight z p q = Pipe $ do
   qstep <- unPipe q
   case qstep of
    Respond z' b k -> return $ Respond (z >> z') b (goRight z p k)
    Pure r         -> z >> return (Pure r)
    Request z' k   -> unPipe $ goLeft z' p k

  goLeft :: Finalizer -> Pipe a b r -> (b -> Pipe b c r) -> Pipe a c r
  goLeft z p q = Pipe $ do
   pstep <- unPipe p
   case pstep of
    Request z' k   -> return $ Request (z >> z') (\a -> goLeft z (k a) q)
    Pure r         -> z >> return (Pure r)
    Respond z' b k -> unPipe $ goRight z' k (q b)

finallyP :: Finalizer -> Pipe a b r -> Pipe a b r
finallyP z p = Pipe $ do
  step <- unPipe p
  case step of
    Request z' k   -> return $ Request (z' >> z) (\a -> finallyP z (k a)) 
    Respond z' b k -> return $ Respond (z' >> z) b (finallyP z k) 
    Pure r         -> z >> return (Pure r)

runPipe :: Pipe () Void r -> IO r 
runPipe p = do
  step <- unPipe p
  case step of
    Pure r        -> return r
    Request _ k   -> runPipe (k ()) 
    Respond _ b _ -> absurd b

prompter :: Pipe a Int r 
prompter = finallyP (putStrLn "</prompter>") $ do
  liftIO $ putStrLn "<prompter>"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: Pipe Int b r
printer = finallyP (putStrLn "</printer>") $ do
  liftIO $ putStrLn "<printer>"
  forever $ do
    i <- request
    liftIO $ print i 

takePipe :: Int -> Pipe a a () 
takePipe n = replicateM_ n $ do
  a <- request
  respond a

main1 :: IO ()
main1 = runPipe (prompter >-> takePipe 3 >-> printer)

-- Note that "Hi" and "Bye" happen when we expect them
-- But we miss a "Bye" on CTRL-C
main2 :: IO ()
main2 = runPipe (forever (prompter >-> takePipe 3) >-> printer)

-- Only one "Hi" and one "Bye" printed
main3 :: IO ()
main3 = runPipe (forever (prompter >-> takePipe 3) >-> takePipe 3 >-> printer)
