{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Local approach to asynchronous exceptions
  In this version we call 'restore' only in the MonadIO instance.
-}
{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE ScopedTypeVariables, RankNTypes #-}

import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Control.Exception
import Data.Void -- from the 'void' package

-- infixl 5 >->

type Restore = forall a. IO a -> IO a

newtype Pipe a b r = Pipe { 
   unPipe :: Restore -> IO (PipeStep a b r) 
 }

type Finalizer = IO ()

data PipeStep a b r =
    Pure (Either SomeException r)
  | Request Finalizer (a -> Pipe a b r)
  | Respond Finalizer b (Pipe a b r) 

instance Monad (Pipe a b) where
  return x = Pipe $ \_ -> return (Pure (Right x))
  x >>= f  = Pipe $ \restore -> do 
               xstep <- unPipe x restore
               case xstep of
                 Request z k    -> return $ Request z (k >=> f)
                 Respond z b k  -> return $ Respond z b (k >>= f)
                 Pure (Right r) -> unPipe (f r) restore
                 Pure (Left e)  -> return $ Pure (Left e)

instance MonadIO (Pipe a b) where 
  liftIO io = Pipe $ \restore -> Pure <$> try (restore io)

respond :: b -> Pipe a b ()
respond b = Pipe $ \_restore -> return $ Respond (return ()) b (return ())

request :: Pipe r b r
request = Pipe $ \_restore -> return $ Request (return ()) return 

-- "Pull based composition"
(>->) :: forall a b c r. Pipe a b r -> Pipe b c r -> Pipe a c r
(>->) = goRight (return ())
  where
    goRight :: Finalizer -> Pipe a b r -> Pipe b c r -> Pipe a c r
    goRight z p q = Pipe $ \restore -> do
      qstep <- unPipe q restore
      case qstep of
        Respond z' b k -> return $ Respond (z >> z') b (goRight z p k)
        Pure r         -> z >> return (Pure r)
        Request z' k   -> unPipe (goLeft z' p k) restore

    goLeft :: Finalizer -> Pipe a b r -> (b -> Pipe b c r) -> Pipe a c r
    goLeft z p q = Pipe $ \restore -> do
      pstep <- unPipe p restore
      case pstep of
        Request z' k   -> return $ Request (z >> z') (\a -> goLeft z (k a) q)
        Pure r         -> z >> return (Pure r)
        Respond z' b k -> unPipe (goRight z' k (q b)) restore

finallyP :: Finalizer -> Pipe a b r -> Pipe a b r
finallyP z p = Pipe $ \restore -> do
  step <- unPipe p restore
  case step of
    Request z' k   -> return $ Request (z' >> z) (\a -> finallyP z (k a)) 
    Respond z' b k -> return $ Respond (z' >> z) b (finallyP z k) 
    Pure r         -> z >> return (Pure r)

runPipe :: Pipe () Void r -> IO r 
runPipe = \p -> mask $ \restore -> go restore p 
  where
    go :: Restore -> Pipe () Void r -> IO r
    go restore p = 
      do step <- unPipe p restore
         case step of
           Pure (Left e)  -> throwIO e
           Pure (Right r) -> return r
           Request _ k    -> go restore (k ()) 
           Respond _ b _  -> absurd b

prompter :: Pipe a Int r 
prompter = finallyP (putStrLn "</prompter>") $ do
  liftIO $ putStrLn "<prompter>"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: Pipe Int b r
printer = finallyP (putStrLn "</printer>") $ do
  liftIO $ putStrLn "<printer>"
  forever $ do
    i <- request
    liftIO $ print i 

takePipe :: Int -> Pipe a a () 
takePipe n = replicateM_ n $ do
  a <- request
  respond a

main1 :: IO ()
main1 = runPipe (forever (prompter >-> takePipe 3) >-> printer)

catchP :: Exception e => Pipe a b r -> (e -> Pipe a b r) -> Pipe a b r
catchP p h = Pipe $ \restore -> do
  step <- unPipe p restore
  case step of
    Pure (Left e)  -> case fromException e of
                        Just e' -> unPipe (h e') restore
                        Nothing -> return $ Pure (Left e)
    Pure (Right r) -> return $ Pure (Right r)
    Request z k    -> return $ Request z (\a -> catchP (k a) h)
    Respond z b k  -> return $ Respond z b (catchP k h) 

exampleHandler1 :: IOException -> Pipe a b ()
exampleHandler1 e = 
  liftIO . putStrLn $ "Whoops! Something terrible happened (" ++ show e ++ ")"

main2 :: IO ()
main2 = runPipe ((forever (prompter >-> takePipe 3) `catchP` exampleHandler1) >-> printer)

exampleHandler2 :: AsyncException -> Pipe a b ()
exampleHandler2 _e = 
  liftIO . putStrLn $ "How dare you interrupt me?"

main3 :: IO ()
main3 = runPipe ((forever (prompter >-> takePipe 3) `catchP` exampleHandler2) >-> printer)

-- Intentionally slow fib
fib :: Int -> Pipe a b Int
fib 0 = return 1
fib 1 = return 1
fib n = do
  n_2 <- fib (n - 2)
  n_1 <- fib (n - 1)
  return (n_2 + n_1)

mapPipeM :: (a -> Pipe a b b) -> Pipe a b r
mapPipeM p = forever $ do
  a <- request
  b <- p a 
  respond b

-- Parse errors are handled correctly, but CTRL_C during fib computation is not
main4 :: IO ()
main4 = runPipe (prompter >-> mapPipeM fib >-> printer) 
