{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Leftovers in conduit style 
-}
{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE ScopedTypeVariables #-}

import Prelude hiding (dropWhile)
import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Data.Void -- from the 'void' package

-- infixl 5 >->
-- infixr 6 >~>

newtype Pipe l a b r = Pipe { unPipe :: IO (PipeStep l a b r) }

data PipeStep l a b r =
    Pure r
  | Request (a -> Pipe l a b r)
  | Respond b (Pipe l a b r) 
  | Leftover l (Pipe l a b r)

instance Monad (Pipe l a b) where
  return x = Pipe $ return (Pure x)
  x >>= f  = Pipe $ do xstep <- unPipe x
                       case xstep of
                         Request k    -> return $ Request (k >=> f)
                         Respond b k  -> return $ Respond b (k >>= f)
                         Pure r       -> unPipe (f r)
                         Leftover l k -> return $ Leftover l (k >>= f)

instance MonadIO (Pipe l a b) where 
  liftIO io = Pipe $ Pure <$> io

respond :: b -> Pipe l a b ()
respond b = Pipe . return $ Respond b (return ())

request :: Pipe l r b r
request = Pipe . return $ Request return 

leftover :: l -> Pipe l a b ()
leftover l = Pipe . return $ Leftover l (return ())

-- "Pull based composition"
(>->) :: forall l a b c r. Pipe l a b r -> Pipe b b c r -> Pipe l a c r
(>->) = goRight
  where
    goRight :: Pipe l a b r -> Pipe b b c r -> Pipe l a c r
    goRight p q = Pipe $ do
      qstep <- unPipe q
      case qstep of
        Respond b k  -> return $ Respond b (goRight p k)
        Pure r       -> return $ Pure r
        Request k    -> unPipe $ goLeft p k
        Leftover l k -> unPipe $ goRight (respond l >> p) k 

    goLeft :: Pipe l a b r -> (b -> Pipe b b c r) -> Pipe l a c r
    goLeft p q = Pipe $ do
      pstep <- unPipe p
      case pstep of
        Request k    -> return $ Request (\a -> goLeft (k a) q)
        Pure r       -> return $ Pure r
        Respond b k  -> unPipe $ goRight k (q b)
        Leftover l k -> return $ Leftover l (goLeft k q)
                      
runPipe :: Pipe Void () Void r -> IO r 
runPipe p = do
  step <- unPipe p
  case step of
    Pure r       -> return r
    Request k    -> runPipe (k ()) 
    Respond b _  -> absurd b
    Leftover l _ -> absurd l

prompter :: Pipe l a Int r 
prompter = do 
  liftIO $ putStrLn "Hi!"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: Pipe l Int b r
printer = forever $ do
  i <- request
  liftIO $ print i 

dropWhile :: (l -> Bool) -> Pipe l l b ()
dropWhile p = do
  x <- request
  if p x
    then dropWhile p
    else leftover x

idPipe :: Pipe a a a r
idPipe = forever $ request >>= respond 

main1 :: IO ()
main1 = runPipe (prompter >-> (dropWhile even >> idPipe) >-> printer)
