{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Pipe transformers 
-}
{-# OPTIONS_GHC -Wall -fno-warn-orphans #-}
{-# LANGUAGE ScopedTypeVariables, FlexibleInstances #-}

import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Data.Void -- from the 'void' package

infixl 5 >->

{------------------------------------------------------------------------------
  Pipe type class
------------------------------------------------------------------------------}

class Pipe p where
  -- Monad 
  returnP :: r -> p a b r 
  bindP   :: p a b r -> (r -> p a b r') -> p a b r'
  -- MonadIO
  liftIOP :: IO r -> p a b r
  -- Pipe specific
  request :: p r b r 
  respond :: b -> p a b ()
  (>->)   :: p a b r -> p b c r -> p a c r

instance Pipe p => Monad (p a b) where
  return = returnP
  (>>=)  = bindP

instance Pipe p => MonadIO (p a b) where
  liftIO = liftIOP

{------------------------------------------------------------------------------
  Basic pipe (cf the Id monad)
------------------------------------------------------------------------------}

newtype IdPipe a b r = Pipe { unPipe :: IO (PipeStep a b r) }

data PipeStep a b r =
    Pure r
  | Request (a -> IdPipe a b r)
  | Respond b (IdPipe a b r) 

instance Pipe IdPipe where
  -- Monad
  returnP x   = Pipe $ return (Pure x)
  x `bindP` f = Pipe $ do xstep <- unPipe x
                          case xstep of
                            Request k   -> return $ Request (k >=> f)
                            Respond b k -> return $ Respond b (k >>= f)
                            Pure r      -> unPipe (f r)

  -- MonadIO
  liftIOP io = Pipe $ Pure <$> io

  -- Pipe
  respond b = Pipe . return $ Respond b (return ())
  request   = Pipe . return $ Request return 
  (>->)     = goRight
    where
      goRight p q = Pipe $ do
        qstep <- unPipe q
        case qstep of
          Respond b k -> return $ Respond b (goRight p k)
          Pure r      -> return $ Pure r
          Request k   -> unPipe $ goLeft p k
  
      goLeft p q = Pipe $ do
        pstep <- unPipe p
        case pstep of
          Request k   -> return $ Request (\a -> goLeft (k a) q)
          Pure r      -> return $ Pure r
          Respond b k -> unPipe $ goRight k (q b)

runIdPipe :: IdPipe () Void r -> IO r 
runIdPipe p = do
  step <- unPipe p
  case step of
    Pure r      -> return r
    Request k   -> runIdPipe (k ()) 
    Respond b _ -> absurd b

{------------------------------------------------------------------------------
  Maybe as a pipe transformer
------------------------------------------------------------------------------}

newtype MaybeP p a b r = MaybeP { unMaybeP :: p a b (Maybe r) }

instance Pipe p => Pipe (MaybeP p) where
  returnP x   = MaybeP $ returnP (Just x)   
  x `bindP` f = MaybeP $ do ma <- unMaybeP x
                            case ma of
                              Nothing -> return Nothing
                              Just a  -> unMaybeP (f a)

  liftIOP io = MaybeP $ Just `liftM` liftIOP io 

  respond b = MaybeP $ Just `liftM` respond b
  request   = MaybeP $ Just `liftM` request
  p >-> q   = MaybeP $ unMaybeP p >-> unMaybeP q

abort :: Pipe p => MaybeP p a b r
abort = MaybeP $ return Nothing

{------------------------------------------------------------------------------
  Examples
------------------------------------------------------------------------------}

prompter :: Pipe p => p a Int r 
prompter = do 
  liftIO $ putStrLn "Hi!"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: Pipe p => p Int b r
printer = forever $ do
  i <- request
  liftIO $ print i 

mapPipe :: Pipe p => (a -> b) -> p a b r
mapPipe f = forever $ do
  a <- request
  respond (f a)

takePipe :: Pipe p => Int -> p a a () 
takePipe n = replicateM_ n $ do
  a <- request
  respond a

abortIf :: Pipe p => (a -> Bool) -> MaybeP p a a r
abortIf p = forever $ do
  a <- request
  if p a then abort
         else respond a

main1 :: IO ()
main1 = runIdPipe $ prompter >-> mapPipe (* 2) >-> takePipe 3 >-> printer

main2 :: IO (Maybe ())
main2 = runIdPipe . unMaybeP $ prompter >-> abortIf even >-> printer
