{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Bidirectional pipes 
-}
{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE ScopedTypeVariables #-}

import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Data.Void -- from the 'void' package

-- infixl 5 >->
-- infixr 6 >~>

newtype Pipe a' a b' b r = Pipe { unPipe :: IO (PipeStep a' a b' b r) }

data PipeStep a' a b' b r =
    Pure r
  | Request a' (a  -> Pipe a' a b' b r)
  | Respond b  (b' -> Pipe a' a b' b r) 

instance Monad (Pipe a' a b' b) where
  return x = Pipe $ return (Pure x)
  x >>= f  = Pipe $ do xstep <- unPipe x
                       case xstep of
                         Request a' k -> return $ Request a' (k >=> f)
                         Respond b  k -> return $ Respond b  (k >=> f)
                         Pure r       -> unPipe (f r)

instance MonadIO (Pipe a' a b' b) where 
  liftIO io = Pipe $ Pure <$> io

respond :: b -> Pipe a' a b' b b' 
respond b  = Pipe . return $ Respond b  return 

request :: a' -> Pipe a' a b' b a
request a' = Pipe . return $ Request a' return 

-- every time we Respond, we get the value of the downstream Request 
-- but this is really the value of the *Next* request 
-- the value of *this* request because an argument to the Pipe

-- "Pull based composition"
(>->) :: (b' -> Pipe a' a b' b r)  
      -> (c' -> Pipe b' b c' c r)
      -> (c' -> Pipe a' a c' c r)
p >-> q = \c' -> p ->> q c'

-- This is approximately the equivalent of goRight in the unidirectional
-- version, but note that there we have (>->) == goRight, while here that
-- is not the case. Instead, the symmetry between (->>) and (>>~) is more
-- evident. 
(->>) :: (b' -> Pipe a' a b' b r)
      ->        Pipe b' b c' c r 
      ->        Pipe a' a c' c r
p ->> q = Pipe $ do
  qstep <- unPipe q 
  case qstep of
    Pure    r    -> return $ Pure r
    Respond c  k -> return $ Respond c (\c' -> p ->> k c') 
    Request b' k -> unPipe $ p b' >>~ k

-- This is the equivalent of goLeft in the unidirectional version
(>>~) ::       Pipe a' a b' b r 
      -> (b -> Pipe b' b c' c r)
      ->       Pipe a' a c' c r 
p >>~ q = Pipe $ do
  pstep <- unPipe p 
  case pstep of
    Pure    r    -> return $ Pure r
    Respond b  k -> unPipe $ k ->> q b
    Request a' k -> return $ Request a' (\a -> k a >>~ q) 

runPipe :: (() -> Pipe Void a b' Void r) -> IO r 
runPipe p = do
  step <- unPipe (p ())
  case step of
    Pure r      -> return r
    Request a' _ -> absurd a' 
    Respond b  _ -> absurd b

{------------------------------------------------------------------------------
  The examples from Pipes4
------------------------------------------------------------------------------}

prompter :: () -> Pipe a' a () Int r 
prompter () = do 
  liftIO $ putStrLn "Hi!"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: () -> Pipe () Int b' b r
printer () = forever $ do
  i <- request ()
  liftIO $ print i 

main1 :: IO ()
main1 = runPipe (prompter >-> printer)

mapPipe :: (a -> b) -> () -> Pipe () a () b r
mapPipe f () = forever $ do
  a <- request ()
  respond (f a)

main2 :: IO ()
main2 = runPipe (prompter >-> mapPipe (* 2) >-> printer)

takePipe :: Int -> () -> Pipe () a () a () 
takePipe n () = replicateM_ n $ do
  a <- request ()
  respond a

main3 :: IO ()
main3 = runPipe (prompter >-> takePipe 3 >-> printer)

{------------------------------------------------------------------------------
  Example with reverse flow 
------------------------------------------------------------------------------}

doubleServer :: Int -> Pipe a' a Int Int r 
doubleServer req = do
  req' <- respond (req * 2)
  doubleServer req'

client :: [Int] -> () -> Pipe Int a b' a () 
client is () = mapM_ aux is
  where
    aux i = do
      resp <- request i
      respond resp

main4 :: IO ()
main4 = runPipe (doubleServer >-> client [1, 2, 3] >-> printer)
